import json
import logging
import os

import requests


def update_resource(fhir_api_url, body, filename, path_info):
    id_resource = None

    type = body["resourceType"]
    id = body["id"]
    url = f"{fhir_api_url}{type}/{id}?"
    # logging.info(f"url: {url}")
    response = requests.put(url, json=body)
    status_code = response.status_code
    if status_code == 200 or status_code == 201:
        logging.info("The resource has successfully been created")
        id_resource = response.json()["id"]
    else:
        logging.info(f"The resource hasn't been created\n" f"{status_code}")
    try:
        with open(path_info) as json_file:
            info = json.load(json_file)
    except FileNotFoundError:
        info = dict()
    info[filename] = id_resource
    with open(path_info, "w") as outfile:
        json.dump(info, outfile)
    return response


def load(path):
    with open(path) as json_file:
        data = json.load(json_file)
    return data


def create_resource_test(path_test, fhir_api_url="http://hapi.fhir.org/baseR4/"):
    path_config = os.path.join(path_test, "infos_test", "config_resources_creation.json")
    config = load(path_config)

    resources_bodies = dict()
    path_resources = os.path.join(path_test, "resources")
    for filename in os.listdir(path_resources):
        body = load(os.path.join(path_resources, filename))
        resources_bodies[filename] = body

    path_info_hapi = os.path.join(path_test, "infos_test", "info_hapi.json")

    for filename in config:
        body = resources_bodies[filename]
        update_resource(fhir_api_url, body, filename, path_info_hapi)


def delete_resource_test(path_test, fhir_api_url="http://hapi.fhir.org/baseR4/"):
    # in config_resources_delete.json is a list of lists with the type of resource
    # and its id to delete
    with open(os.path.join(path_test, "infos_test", "config_resources_delete.json")) as json_file:
        list_to_delete = json.load(json_file)
        logging.info(f"list to delete : {list_to_delete}")

    for resource in list_to_delete:
        type_resource = resource[0]
        id_resource = resource[1]
        url = f"{fhir_api_url}{type_resource}/{id_resource}"
        response = requests.delete(url)
        status_code = response.status_code
        if status_code == 200 or status_code == 204 or status_code == 202:
            logging.info("The resource has successfully been delete")
        else:
            logging.info(f"The resource hasn't been delete\n" f"{status_code}")



import logging

import pandas as pd
import pytest

import fhir2dataset as query
from fhir2dataset.api import ApiCall, ApiRequest, BearerAuth, Response
from fhir2dataset.data_class import Element, Elements


def test_api_call_get_response():
    url = "http://hapi.fhir.org/baseR4/Patient"
    api_call = ApiCall(url=url)
    response = api_call._get_response(url)

    assert isinstance(response, Response)
    assert "http://hapi.fhir.org/baseR4" in response.next_url

    url = "http://hapi.fhir.org/baseR4/Demon"
    with pytest.raises(ValueError):
        api_call._get_response(url)

    url = "http://hapi.fhir.org/baseR4/Patient?lolz=1"
    with pytest.raises(ValueError):
        api_call._get_response(url)

    url = "http://hapi.fhir.org/baseR1/Patient"
    with pytest.raises(ValueError):
        api_call._get_response(url)


def test_api_call_get_count():
    url = "http://hapi.fhir.org/baseR4/Patient"
    api_call = ApiCall(url=url)
    total = api_call._get_count(url)
    assert isinstance(total, int)
    assert total > 0


def test_api_call_fix_next_url():
    url = "http://hapi.fhir.org/baseR4/Patient"
    api_call = ApiCall(url=url)
    fixed_url = api_call._fix_next_url(url)

    assert "_count" in fixed_url


@pytest.mark.parametrize(
    "bad_url, good_url",
    [
        (
            "http://arkhn.org/Patient?gender=male?param=value",
            "http://arkhn.org/Patient?gender=male&param=value",
        ),
        ("http://arkhn.org/Patient/?gender=male", "http://arkhn.org/Patient?gender=male"),
    ],
)
def test_api_fix_url(bad_url, good_url):
    fix_url = ApiCall._ApiCall__fix_url
    assert fix_url(bad_url) == good_url


def test_api_request_get_all():
    """
    Test that the next link is used correctly and that more than 1 page is fetched
    """
    PAGE_SIZE = query.api.PAGE_SIZE

    query.api.PAGE_SIZE = 30
    url = "http://hapi.fhir.org/baseR4/Patient?birthdate=2000-01-01"
    elements = Elements()
    elements.append(Element("gender", "Patient.gender"))
    call_api = ApiRequest(url, elements)
    results = call_api.get_all()

    if len(call_api.df) < query.api.PAGE_SIZE:
        logging.warn(
            f"test_get_all couldn't be tested properly. Only {len(call_api.df)} "
            f"results returned for page size {query.api.PAGE_SIZE}."
        )

    assert len(results) > query.api.PAGE_SIZE

    # restore the original PAGE_SIZE
    query.api.PAGE_SIZE = PAGE_SIZE


def test_api_request__get_data():
    url = "http://hapi.fhir.org/baseR4/Patient?birthdate=2000-01-01"
    elements = Elements()
    elements.append(Element("birthdate", "Patient.birthDate"))
    call_api = ApiRequest(url, elements)

    results = [
        {
            "resource": {
                "resourceType": "Patient",
                "id": "23",
                "birthDate": "2000-01-01",
                "gender": "male",
            }
        },
        {
            "resource": {
                "resourceType": "Patient",
                "id": "24",
                "birthDate": "2000-01-01",
                "gender": "male",
            }
        },
    ]
    expected_tabular_results = pd.DataFrame({"birthdate": ["2000-01-01", "2000-01-01"]})

    tabular_results = call_api._get_data(results)

    assert (expected_tabular_results == tabular_results).all().bool()



import logging

import pandas as pd
import tqdm

from fhir2dataset.api import ApiRequest
from fhir2dataset.fhirrules import FHIRRules
from fhir2dataset.graphquery import GraphQuery
from fhir2dataset.tools.graph import join_path
from fhir2dataset.url_builder import URLBuilder

logger = logging.getLogger(__name__)


class Query:
    """Query Executor

    Description:
        This module purpose is to perform SQL-type queries whose information is
        filled in configuration files of the form:
        ```
        {
            "from": {
                "alias_1": "Resource type 1",
                "alias_2": "Resource type 2",
                "alias_3": "Resource type 3",
                ...
            },
            "select": {
                "alias_1": [
                    "searchparam/fhirpath of attribute a of resource type 1",
                    "searchparam/fhirpath of attribute b of resource type 1",
                    "searchparam/fhirpath of attribute c of resource type 1"
                ],
                "alias_2": [
                    "searchparam of attribute a of resource type 2",
                    ...
                ],
                ...
            },
            "join": {
                "inner": {
                    "alias_1": {
                        "searchparam of attribute d, which is of type Reference, of resource type 1": "alias_2"
                    },
                    ...
                },
                "inner": {
                    "alias_2": {
                        "searchparam of attribute b, which is of type Reference, of resource type 2": "alias_3"
                    },
                    ...
                }
            },
            "where": {
                "alias_2": {
                    "searchparam of attribute c of resource type 2": "value 1",
                    "searchparam of attribute d of resource type 2": "value 2"
                },
                "alias_3": {
                    "searchparam of attribute a of resource type 2": "value 3",
                    "searchparam of attribute b of resource type 2": "value 4"
                },
                ...
            }
        }
        ```
        for the next associated SQL query:
        ```
        SELECT alias_1.a, alias_1.b, alias_1.c, alias_2.a FROM (Resource type 1) as alias_1
        INNER JOIN (Resource type 2) as alias_2
        ON alias_1.d = alias_2.id
        INNER JOIN (Resource type 3) as alias_3
        ON alias_2.b = alias_3.id
        WHERE alias_2.c = "value 1"
        AND alias_2.d = "value 2"
        AND alias_3.a = "value 3"
        AND alias_3.b = "value 4"
        ```

    Attributes:
        config (dict): dictionary storing the initial request
        graph_query (GraphQuery): instance of a GraphQuery object that gives a graphical
            representation of the query
        dataframes (dict): dictionary storing for each alias the resources requested on
            the api in tabular format
        main_dataframe (DataFrame): pandas dataframe storing the final result table
    """  # noqa

    def __init__(
        self,
        fhir_api_url: str = None,
        token: str = None,
        fhir_rules: FHIRRules = None,
    ):
        """Requestor's initialisation

        Arguments:
            fhir_api_url (str): the service base URL (e.g. http://hapi.fhir.org/baseR4/)
            token (str): (Optional) the bearer token to authenticate to the FHIR server,
                if necessary
            fhir_rules (FHIRRules): (Optional) an instance of FHIR rules, initialized
                with search parameters
        """  # noqa
        self.fhir_api_url = fhir_api_url or "http://hapi.fhir.org/baseR4/"
        if not fhir_rules:
            fhir_rules = FHIRRules(fhir_api_url=self.fhir_api_url)
        self.fhir_rules = fhir_rules
        self.token = token

        self.config = None
        self.graph_query = None
        self.dataframes = {}
        self.main_dataframe = None

    def from_config(self, config: dict):
        """Executes the query from a dictionary in the format of a configuration file

        Arguments:
            config (dict): dictionary in the format of a configuration file
        """
        self.config = {
            "from_dict": config.get("from", None),
            "select_dict": config.get("select", None),
            "where_dict": config.get("where", None),
            "join_dict": config.get("join", None),
        }
        return self

    def execute(self, debug: bool = False):
        """Executes the complete query

        1. constructs a GraphQuery object to store the query as a graph
        2. builds the url of the requests to send to the API
        3. retrieves the answers from the api and puts them as tables in the dataframes attribute
        4. executes joins to return the result table in the main_dataframe attribute

        Arguments:
            debug (bool): if debug is true then the columns needed for internal processing
                are kept in the final dataframe. Otherwise only the columns of the select are
                kept in the final dataframe. (default: {False})
        """  # noqa
        self.graph_query = GraphQuery(fhir_api_url=self.fhir_api_url, fhir_rules=self.fhir_rules)
        self.graph_query.build(**self.config)

        with tqdm.tqdm(
            total=1, unit_scale=100, bar_format="{l_bar}{bar}| {n:.02f}/{total:.02f}"
        ) as pbar:
            bar_frac = 1 / len(self.graph_query.resources_by_alias)
            for resource_alias, resource in self.graph_query.resources_by_alias.items():
                url = URLBuilder(
                    fhir_api_url=self.fhir_api_url,
                    graph_query=self.graph_query,
                    main_resource_alias=resource_alias,
                ).compute()
                call = ApiRequest(
                    url=url,
                    elements=resource.elements,
                    token=self.token,
                    pbar=pbar,
                    bar_frac=bar_frac,
                )
                self.dataframes[resource_alias] = call.get_all()

        self._clean_columns()

        for resource_alias, dataframe in self.dataframes.items():
            logger.debug(f"{resource_alias} dataframe builded head - \n{dataframe.to_string()}")

        # We check if there is more than 1 alias of resource
        if len(self.dataframes) > 1:
            self.main_dataframe = self._join()
        else:
            self.main_dataframe = list(self.dataframes.values())[0]

        self.main_dataframe = self.main_dataframe.reset_index(drop=True)

        logger.debug(
            f"Main dataframe builded head before columns selection-"
            f"\n{self.main_dataframe.to_string()}"
        )
        if not debug:
            self.__select_columns()
            self.__remove_lists()

        return self.main_dataframe

    def _join(self) -> pd.DataFrame:
        """Execute the joins one after the other in the order specified by the
        join_path function.

        Returns:
            pd.DataFrame: dataframe containing all joined resources
        """
        list_join = join_path(self.graph_query.resources_graph)
        main_alias_join = list_join[0][0]
        main_df = self.dataframes[main_alias_join]
        for alias_1, alias_2 in list_join:
            df_1 = main_df
            df_2 = self.dataframes[alias_2]
            main_df = self._join_2_df(alias_1, alias_2, df_1, df_2)
        return main_df

    def _join_2_df(
        self, alias_1: str, alias_2: str, df_1: pd.DataFrame, df_2: pd.DataFrame
    ) -> pd.DataFrame:
        """Executes the join between two dataframes

        The join key is the id of the child resource.
        This id is contained in :
            * in the column child_alias:id of the child resource table named child_alias
              (e.g. parent:id for a parent dataframe)
            * in the alias_parent:searchparam_parent column of the parent resource table
               named alias_parent (e.g. condition:subject.reference for a condition dataframe)

        The function is in charge of finding out who is the mother resource and who is the
        daughter resource.

        Arguments:
            alias_1 (str): df_1 alias
            alias_2 (str): df_2 alias
            df_1 (pd.DataFrame): dataframe containing the elements of a resource
            df_2 (pd.DataFrame): dataframe containing the elements of a resource
        Returns:
            pd.DataFrame: dataframe containing the elements of the 2 resources according to
                an inner join
        """  # noqa
        edge_info = self.graph_query.resources_graph.edges[alias_1, alias_2]["info"]
        alias_parent = edge_info.parent
        alias_child = edge_info.child

        how = edge_info.join_how
        searchparam_parent = edge_info.searchparam_parent

        parent_on = f"{alias_parent}:join_{searchparam_parent}"
        child_on = f"{alias_child}:from_id"

        if how == "child":
            how = "right"
        elif how == "parent":
            how = "left"
        else:
            how = "inner"

        if alias_1 == alias_parent:
            df_1 = df_1.explode(parent_on)
            df_2 = df_2.explode(child_on)
            df_merged_inner = pd.merge(
                left=df_1, right=df_2, left_on=parent_on, right_on=child_on, how=how
            )
        else:
            df_1 = df_1.explode(child_on)
            df_2 = df_2.explode(parent_on)
            df_merged_inner = pd.merge(
                left=df_2, right=df_1, left_on=parent_on, right_on=child_on, how=how
            )
        return df_merged_inner

    def _clean_columns(self):
        """Perform preprocessing on all dataframes harvested in the dataframe attribute:
        - Add the resource type in front of an element id so that the resource id matches
          the references of its parent resource references
        - Add the table alias as a prefix to each column name
        """  # noqa

        def _add_resource_type_to_id(df, resource_type: str):
            # add assert to check that there is only one id in list
            df["from_id"] = f"{resource_type}/" + df["from_id"]
            return df

        for resource_alias, df in self.dataframes.items():
            resource_type = self.graph_query.resources_by_alias[resource_alias].resource_type

            df = df.pipe(_add_resource_type_to_id, resource_type=resource_type)

            self.dataframes[resource_alias] = df.add_prefix(f"{resource_alias}:")

    def __select_columns(self):
        """Clean the final dataframe to keep only the columns of the select"""
        final_columns = []
        for resource_alias, resource in self.graph_query.resources_by_alias.items():
            for element in resource.elements.where(goal="select"):
                final_columns.append(f"{resource_alias}:{element.col_name}")
        self.main_dataframe = self.main_dataframe[final_columns]

    def __remove_lists(self):
        """Remove lists from columns with only single elements"""

        def unlist(x):
            """Auxiliary function to squeeze lists with one element"""
            if isinstance(x, list) and len(x) == 1:
                return unlist(x[0])
            else:
                return x

        if len(self.main_dataframe) > 0:
            df = self.main_dataframe.copy()
            for column in df.columns:
                df[column] = df[column].apply(unlist)
            self.main_dataframe = df



from fhir2dataset.graphquery import GraphQuery
from fhir2dataset.parser import Parser
from fhir2dataset.query import Query


def test_graphquery():
    sql_query = """
    SELECT
    Patient.identifier.type.coding,
    Patient.name.family,
    Patient.birthDate
    FROM Patient
    INNER JOIN Practitioner
    ON Patient.general-practitioner = Practitioner.id
    WHERE Patient.birthdate=ge2001-01-01
    """

    config = Parser().from_sql(sql_query)
    query = Query().from_config(config)
    graph_query = GraphQuery(fhir_api_url=query.fhir_api_url, fhir_rules=query.fhir_rules)
    graph_query.build(**query.config)
    graph = graph_query.resources_by_alias

    assert len(graph) == 2

    assert len(graph["Patient"].elements.elements) == 6
    assert len(graph["Practitioner"].elements.elements) == 1

    assert len(graph["Patient"].elements.where(goal="join")) == 1



import json
import logging
import os

import pytest

from fhir2dataset.query import Query
from tests.tools import delete_resource_test

log_format = "[%(asctime)s] [%(levelname)s] - %(message)s"
logging.basicConfig(level=logging.INFO, format=log_format)


@pytest.mark.parametrize(
    "dirname, fhir_api_url",
    [("tests/empty", "http://hapi.fhir.org/baseR4/")],
)
def test_dataframe_empty(dirname, fhir_api_url):
    delete_resource_test(dirname, fhir_api_url=fhir_api_url)
    with open(os.path.join(dirname, "config.json")) as json_file:
        config = json.load(json_file)

    query = Query(fhir_api_url)
    query.from_config(config)
    query.execute(debug=True)
    df = query.main_dataframe

    assert df.empty



import random

import matplotlib.pyplot as plt
import networkx as nx

OPEN_CHAR = ["[", "{", "("]
CLOSE_CHAR = ["]", "}", ")"]
SAFE_CHAR = ["'", '"']
DEFAULT_BREAKLINE_CHAR = "\n"
DEFAULT_INDENT_CHAR = " "


def draw_graphquery(resources_alias_graph):
    """draws the resources_graph attribute"""
    import matplotlib.pyplot as plt

    edge_labels = {}
    for i in resources_alias_graph.edges(data=True):
        edge_infos = custom_repr(i[2]["info"].__repr__())
        edge_labels[i[0:2]] = edge_infos

    plt.figure(figsize=(15, 15))
    layout = nx.spring_layout(resources_alias_graph)
    nx.draw_networkx(resources_alias_graph, pos=layout)
    nx.draw_networkx_labels(resources_alias_graph, pos=layout)
    nx.draw_networkx_edge_labels(
        resources_alias_graph,
        pos=layout,
        edge_labels=edge_labels,
        font_size=10,
        rotate=False,
        horizontalalignment="left",
    )
    plt.show()


def custom_repr(
    string: str,
    indent_width: int = 4,
    breakline_char: str = DEFAULT_BREAKLINE_CHAR,
    indent_char: str = DEFAULT_INDENT_CHAR,
):
    nested_int = 0
    new_string = ""
    safe_zone = False
    len_breakline_char = len(breakline_char)
    for character in string:
        if character in SAFE_CHAR:
            safe_zone = not safe_zone
            new_string += character
        elif safe_zone:
            new_string += character
        elif character in OPEN_CHAR:
            nested_int += 1
            new_string += character + breakline_char + nested_int * indent_width * indent_char
        elif character == ",":
            if new_string[-len_breakline_char:] == breakline_char:
                new_string = new_string[:-len_breakline_char]
            new_string += "," + breakline_char + (nested_int * (indent_width) - 1) * indent_char
        elif character in CLOSE_CHAR:
            nested_int -= 1
            if new_string[-len_breakline_char:] == breakline_char:
                new_string = new_string[:-len_breakline_char]
            new_string += (
                breakline_char
                + nested_int * indent_width * indent_char
                + character
                + breakline_char
            )
        else:
            new_string += character
    return new_string


def hierarchy_pos(G, root=None, width=1.0, vert_gap=0.2, vert_loc=0, xcenter=0.5):  # noqa
    """
    From Joel's answer at https://stackoverflow.com/a/29597209/2966723.
    Licensed under Creative Commons Attribution-Share Alike

    If the graph is a tree this will return the positions to plot this in a
    hierarchical layout.

    Arguments:
        G: the graph (must be a tree)
        root: the root node of current branch
            - if the tree is directed and this is not given,
              the root will be found and used
            - if the tree is directed and this is given, then
              the positions will be just for the descendants of this node.
            - if the tree is undirected and not given,
              then a random choice will be used.
        width: horizontal space allocated for this branch - avoids overlap with other branches
        vert_gap: gap between levels of hierarchy
        vert_loc: vertical location of root
        xcenter: horizontal location of root
    """
    if not nx.is_tree(G):
        raise TypeError("cannot use hierarchy_pos on a graph that is not a tree")

    if root is None:
        if isinstance(G, nx.DiGraph):
            root = next(
                iter(nx.topological_sort(G))
            )  # allows back compatibility with nx version 1.11
        else:
            root = random.choice(list(G.nodes))

    def _hierarchy_pos(
        G, root, width=1.0, vert_gap=0.2, vert_loc=0, xcenter=0.5, pos=None, parent=None  # noqa
    ):
        """
        see hierarchy_pos docstring for most arguments

        pos: a dict saying where all nodes go if they have been assigned
        parent: parent of this branch. - only affects it if non-directed

        """

        if pos is None:
            pos = {root: (xcenter, vert_loc)}
        else:
            pos[root] = (xcenter, vert_loc)
        children = list(G.neighbors(root))
        if not isinstance(G, nx.DiGraph) and parent is not None:
            children.remove(parent)
        if len(children) != 0:
            dx = width / len(children)
            nextx = xcenter - width / 2 - dx / 2
            for child in children:
                nextx += dx
                pos = _hierarchy_pos(
                    G,
                    child,
                    width=dx,
                    vert_gap=vert_gap,
                    vert_loc=vert_loc - vert_gap,
                    xcenter=nextx,
                    pos=pos,
                    parent=root,
                )
        return pos

    return _hierarchy_pos(G, root, width, vert_gap, vert_loc, xcenter)


def show_tree(graph, number=None):
    """
    Function used to plot the graph query
    """
    plt.figure(figsize=(5, 5))
    layout = hierarchy_pos(graph)
    labels = nx.get_node_attributes(graph, "column_idx")
    for key, column_idx in labels.items():
        labels[key] = f"{key.fhirpath}\n{column_idx}"
    nx.draw_networkx(graph, labels=labels, pos=layout)
    plt.show()



# FIXME: This file needs FHIR2Dataset#96 to be used

import logging
from dataclasses import asdict, dataclass
from typing import List

import networkx as nx

from fhir2dataset.tools.fhirpath import parse_fhirpath

logger = logging.getLogger(__name__)


@dataclass(eq=True, frozen=True)
class Node:
    """Modeling a node of a process tree"""

    fhirpath: str
    index: str
    previous_node_hash: int = 0


def split_fhirpath(fhirpath: str) -> List[Node]:
    """Transforms a fhirpath into a list of sub-fhirpaths that run one after the other would give
    the same result as the initial fhirpath.

    Args:
        fhirpath (str): a string of characters representing a fhirpath

    Returns:
        List[Node]: a list of Nodes representing the sub-fhirpaths described above

    Examples:
        1. "Patient.name.given" becomes ["Patient","name","given"]
        1. "Patient.name | Practitioner.name" becomes ["Patient.name | Practitioner.name"]
        2. "Patient.(x | y).use" becomes ["Patient","(x | y)","use"]
    """

    def _create_node_list(sub_fhirpaths: List[str]):
        node_list = []
        node = None
        for idx, fhirpath in enumerate(sub_fhirpaths):
            if node:
                node = Node(fhirpath=fhirpath, index=str(idx), previous_node_hash=hash(node))
            else:
                node = Node(fhirpath=fhirpath, index=str(idx))
            node_list.append(node)
        return node_list

    sub_paths = [sub_path for sub_path in fhirpath.split(".") if sub_path != ""]

    tmps_list = []
    tmp_word = ""
    num_brackets = 0
    num_or = fhirpath.count("|")

    for sub_path in sub_paths:
        tmp_word = f"{tmp_word}.{sub_path}" if tmp_word else sub_path
        for character in sub_path:
            if character == "(":
                num_brackets += 1
            elif character == ")":
                num_brackets -= 1
            elif character == "|" and num_brackets % 2 == 1:
                num_or -= 1
        if not num_brackets:
            tmps_list.append(tmp_word)
            tmp_word = ""
    if num_or != 0:
        tmps_list = [fhirpath]

    node_list = _create_node_list(tmps_list)
    logger.debug(f"the fhirpath: {fhirpath} is parsed in {[node.fhirpath for node in node_list]}")
    return node_list


class Forest:
    """class modeling the entire set of process trees

    Attributes:
        trees (dict): dictionary where the key is the value of the root and the value of the
        associated complete tree
        num_exp (int): total number of fhirpath that will be computed by this process tree forest
    """

    def __init__(self):
        self.trees = {}
        self.num_exp = 0

    def add_fhirpath(self, fhirpath: str) -> None:
        """Adding a fhirpath to the forest is done in the following steps:
            1. the fhirpath is divided into sub-fhirpaths which executed one after the other give
            the same result as the execution of the whole fhirpath.
            2. these subfhirpaths are added to the forest, each subfhirpath representing a node of
            a tree. The interest is that if 2 fhirpaths share the same first sub-fhirpaths then they
            will share the same first nodes of the tree.

        Args:
            fhirpath (str): string representing a fhirpath
        """
        splitted_fhirpath = split_fhirpath(fhirpath)
        self.__add_splitted_fhirpath(splitted_fhirpath)

    def simplify_trees(self):
        """This function, executed once all the fhirpaths have been added, allows to reduce the
        trees in the forest. If a node has only one successor node and if they are tagged with
        exactly the same column numbers (the same processes), then they will be merged.
        """
        new_trees = {}
        for root, tree in self.trees.items():
            tree.simplify_tree()
            new_trees[tree.root] = tree
        self.trees = new_trees

    def parse_fhirpaths(self):
        """transforms each expression that corresponds to a sub fhirpath of a node into a parsed
        version used by the fhirpath.js library.
        """
        for tree in self.trees.values():
            tree.parse_fhirpaths()

    def create_forest_dict(self):
        """creates a dictionary modeling the forest understandable by the javascript function in
        the forest.js file

        Returns:
            dict: dictionary described above
        """
        forest_dict = {}
        for root, tree in self.trees.items():
            nodes = tree.graph.nodes
            edges = tree.graph.edges

            nodes_dict = {}
            for node in nodes:
                nodes_dict[str(hash(node))] = asdict(node)
                nodes_dict[str(hash(node))]["column_idx"] = nodes[node]["column_idx"]
                nodes_dict[str(hash(node))]["parsed_fhirpath"] = nodes[node]["parsed_fhirpath"]

            edges_array = []
            for edge in edges:
                edges_array.append([str(hash(edge[0])), str(hash(edge[1]))])

            forest_dict[str(hash(root))] = {"nodes_dict": nodes_dict, "edges_array": edges_array}
        return forest_dict

    def __add_splitted_fhirpath(self, splitted_fhirpath=List[Node]):
        root = splitted_fhirpath[0]
        if root not in self.trees.keys():
            self.trees[root] = Tree(root)
        self.trees[root].add_edges_from_splitted_fhirpath(splitted_fhirpath, self.num_exp)
        self.num_exp += 1


class Tree:
    """class modeling a process tree

    Attributes:
        root (Node): the root of the tree
        graph (nx.DiGraph): oriented graph representing the tree
    """

    def __init__(self, root: Node):
        self.root = root
        self.graph = nx.DiGraph()

    def add_edges_from_splitted_fhirpath(self, splitted_fhirpath: list, column_idx: int):
        """add splitted_fhirpath to the tree, each sub-part of the fhirpath representing a node of
        the tree. The interest is that if 2 fhirpaths share the same first sub-fhirpaths then they
        will share the same first nodes of the tree

        Args:
            splitted_fhirpath (List[str]): list of fhirpaths representing a process
            column_idx ([type]): column number corresponding to the process number
        """
        for node in splitted_fhirpath:
            if node in self.graph.nodes:
                self.graph.nodes[node]["column_idx"].append(column_idx)
            else:
                self.graph.add_node(node, column_idx=[column_idx])
        edges = []
        for idx in range(len(splitted_fhirpath) - 1):
            edges.append((splitted_fhirpath[idx], splitted_fhirpath[idx + 1]))
        self.graph.add_edges_from(edges)

    def parse_fhirpaths(self):
        """transforms each expression that corresponds to a sub fhirpath of a node into a parsed
        version used by the fhirpath.js library.
        """
        for node in nx.dfs_preorder_nodes(self.graph, self.root):
            self.graph.nodes[node]["parsed_fhirpath"] = parse_fhirpath(node.fhirpath)

    def simplify_tree(self):
        """This function, executed once all the fhirpaths have been added, allows to reduce the
        tree. If a node has only one successor node and if they are tagged with exactly the same
        column numbers (the same processes), then they will be merged.
        """
        tmp_root = self.root
        final_root = self.root
        new_graph = self.graph.copy()

        previous_node_new_graph = None
        previous_node_created_new_graph = None
        column_idx = nx.get_node_attributes(self.graph, "column_idx")

        for node_old_graph in nx.dfs_preorder_nodes(self.graph, self.root):
            successors_old_graph = list(self.graph.successors(node_old_graph))
            if len(successors_old_graph) == 1:
                successor_old_graph = successors_old_graph[0]
                if column_idx[successor_old_graph] == column_idx[node_old_graph]:
                    if (
                        previous_node_created_new_graph
                        and column_idx[successor_old_graph] == column_idx[node_old_graph]
                    ):
                        previous_node_new_graph = previous_node_created_new_graph
                    else:
                        previous_node_new_graph = node_old_graph

                    new_fhirpath = (
                        f"{previous_node_new_graph.fhirpath}.{successor_old_graph.fhirpath}"
                    )
                    new_index = f"{previous_node_new_graph.index}-{successor_old_graph.index}"
                    new_node_new_graph = Node(new_fhirpath, new_index)

                    nx.relabel_nodes(
                        new_graph, {successor_old_graph: new_node_new_graph}, copy=False
                    )
                    new_graph = nx.contracted_nodes(
                        new_graph, new_node_new_graph, previous_node_new_graph, self_loops=False
                    )
                    previous_node_created_new_graph = new_node_new_graph
                    if node_old_graph == tmp_root:
                        tmp_root = successor_old_graph
                        final_root = new_node_new_graph
            else:
                previous_node_created_new_graph = None
        self.graph = new_graph
        self.root = final_root



import json
import logging
import os
from collections import defaultdict
from functools import lru_cache
from typing import List

from fhir2dataset.data_class import SearchParameter

logger = logging.getLogger(__name__)

DEFAULT_METADATA_DIR = "tools/metadata"


class SearchParameters:
    def __init__(self, search_parameters: List[SearchParameter] = None):
        self.items = search_parameters or []
        self._data = defaultdict(lambda: defaultdict(dict))

        for search_parameter in self.items:
            self._add_data(search_parameter)

    def add(self, search_parameters):
        if isinstance(search_parameters, SearchParameter):
            self.items.append(search_parameters)
            self._add_data(search_parameters)
        elif isinstance(search_parameters, list):
            for search_parameter in search_parameters:
                self.add(search_parameter)
        else:
            raise TypeError(
                f"{search_parameters} should be a list or a SearchParameter type instead of "
                f"{type(search_parameters)} type"
            )

    def searchparam_to_fhirpath(self, search_param: str, resource_type: str = "all"):
        """Retrieve the fhirpath associated to a searchparam of a certain resource type

        Arguments:
            resource_type (str): name of a resource type (e.g. 'Organization')
            search_param (str): searchparam of this resource type (e.g. 'address-postalcode')

        Returns:
            str: the fhirpath associated to the searchparam (e.g. 'address.postalCode')
        """  # noqa
        fhirpath = self._data[search_param][resource_type]

        if fhirpath == {}:  # self._data is a defaultdict of dict: so it never KeyErrors
            logger.info(f"The searchparam '{search_param}' doesn't exist in the rules")
            return None

        return fhirpath

    def _add_data(self, search_parameter: SearchParameter):
        """Fill the _data store dict with information from a searchparam"""
        fhirpath = search_parameter.fhirpath
        code = search_parameter.code
        resource_types = search_parameter.resource_types

        for resource_type in resource_types:
            self._data[code][resource_type] = fhirpath


class FHIRRules:
    """Class storing rules specific to the FHIR syntax and/or the FHIR API used,
    such as the search parameters

    Attributes:
        searchparameters (SearchParameters): an instance json of a SearchParameters resource
    """  # noqa

    def __init__(
        self,
        fhir_api_url: str = None,
        path: str = None,
        searchparameters_filename: str = "SearchParameters.json",
    ):
        """
        Arguments:
            fhir_api_url (str): The Service Base URL (e.g. http://hapi.fhir.org/baseR4/)
                (default: {None})
            path (str): path to the folder containing the searchparameters file (default: None)
            searchparameters_filename (str): filename of a json that contains a resource of
                type SearchParameters  (default: {"SearchParameters.json"})
        """  # noqa
        self.fhir_api_url = fhir_api_url
        self.path = path or os.path.join(os.path.dirname(__file__), DEFAULT_METADATA_DIR)
        self.searchparameters_filename = searchparameters_filename
        self.searchparameters = self.build_searchparameters()

    @lru_cache(maxsize=10000)
    def searchparam_to_fhirpath(self, search_param: str, resource_type: str = "all"):
        """Retrieve the fhirpath associated to a searchparam of a certain resource type,
        and filter it to only clauses that are associated to the given resource_type

        Arguments:
            resource_type (str): name of a resource type (e.g. 'Organization')
            search_param (str): searchparam of this resource type (e.g. 'address-postalcode')

        Returns:
            str: the fhirpath associated to the searchparam (e.g. 'address.postalCode')
        """  # noqa
        fhirpath = self.searchparameters.searchparam_to_fhirpath(search_param, resource_type)

        if fhirpath and resource_type != "all":
            params = fhirpath.split(" | ")
            filtered_params = []
            for param in params:
                if resource_type in param:
                    filtered_params.append(param)
            fhirpath = " | ".join(filtered_params)

            if len(filtered_params) == 0:
                return ValueError(
                    f"There was an error while filtrating {fhirpath} on "
                    f"resource {resource_type}"
                )
        return fhirpath

    def build_searchparameters(self) -> SearchParameters:
        """builds an instance of SearchParameters storing all the possible searchparameters
        (instance of SearchParameter) whose information comes from a bundle composed only of
        FHIR resources of type SearchParameter according to the following process:

        1. Retrieves information of interest in each instance of the bundle. For each instance
            of the bundle are retrieved: the code, the expression (=fhirpath) and the base
            (=the types of resources on which these fhirpaths can be applied)
        2. With these 3 pieces of information a SearchParameter object is created
        3. SearchParameter objects are stored in a SearchParameters object.

        Returns:
            SearchParameters: instance described above
        """  # noqa
        bundle = self._get_from_file(self.path, self.searchparameters_filename)

        resources = [resource["resource"] for resource in bundle["entry"]]

        search_parameters = SearchParameters()
        for resource in resources:
            if "expression" in resource:
                search_parameters.add(
                    SearchParameter(
                        code=resource["code"],
                        fhirpath=resource["expression"],
                        resource_types=resource["base"],
                    )
                )

        return search_parameters

    def _get_from_file(self, path: str, filename: str) -> dict:
        """Get a json (dict) from a file

        Arguments:
            path (str): path to the folder containing the filename
            filename (str): filename of the json

        Returns:
            dict: dict containing the json
        """
        with open(os.path.join(path, filename)) as json_file:
            file_dict = json.load(json_file)
        return file_dict



import re

import pandas as pd

from fhir2dataset.fhirrules import FHIRRules  # noqa
from fhir2dataset.parser import Parser  # noqa
from fhir2dataset.query import Query  # noqa


def sql(sql_query: str, fhir_api_url: str = None, token: str = None) -> pd.DataFrame:
    """Interpret a SQL-like query and query a FHIR api

    Arguments:
        sql_query (str): A query in a SQL-like syntax
        fhir_api_url (str): the base url of the FHIR server (e.g. http://hapi.fhir.org/baseR4/)
        token (str): a Bearer Auth token

    Returns:
        pd.Dataframe: the result of the query in a tabular format
    """
    config = Parser().from_sql(sql_query)
    query = Query(fhir_api_url=fhir_api_url, token=token).from_config(config)
    df = query.execute()
    # rename the columns to match the sql syntax
    # patient:Patient.name.given -> patient.name.given
    df = df.rename(lambda x: re.sub("\:\w+\.", ".", x), axis="columns")  # noqa
    return df



import logging
from collections import defaultdict
from posixpath import join as urljoin
from typing import Type

from fhir2dataset.data_class import SearchParameter
from fhir2dataset.graphquery import GraphQuery

logger = logging.getLogger(__name__)


class URLBuilder:
    """class that allows to build the url of a query that will be made for an alias
    of a FHIR resource to a FHIR api

    Attributes:
        fhir_api_url (str): The Service Base URL (e.g. http://hapi.fhir.org/baseR4/)
        graph_query (GraphQuery): instance of a GraphQuery object that gives a graphical
            representation of the global query
        main_resource_alias (str): alias given to a set of fhir resources of a certain type
            which are the subject of the api query
    """  # noqa

    def __init__(
        self, fhir_api_url: str, graph_query: GraphQuery, main_resource_alias: str
    ) -> None:
        """
        Arguments:
            fhir_api_url (str): The Service Base URL (e.g. http://hapi.fhir.org/baseR4/)
            graph_query (GraphQuery): instance of a GraphQuery object that gives a graphical
                representation of the global query
            main_resource_alias (str): alias given to a set of fhir resources of a certain type
        """  # noqa
        self.fhir_api_url = fhir_api_url
        self.graph_query = graph_query
        self.main_resource_alias = main_resource_alias

        self._params = defaultdict(list)

    def compute(self):
        self._update_url_params()
        return self._compute_url()

    def _compute_url(self) -> str:
        """Generates the url which will make it possible to recover the resources of the type
        of main_alias respecting as well as possible the conditions where on itself and on its
        neighbors.

        Returns:
            str: corresponding API request url
        """  # noqa
        params = f"{self.graph_query.resources_by_alias[self.main_resource_alias].resource_type}?"

        for key, values in self._params.items():
            for value in values:
                params += f"{key}={value}&"

        # Rm last '&'
        params = params[:-1]

        search_query_url = urljoin(self.fhir_api_url, params)
        logger.info(f"the computed url is {search_query_url}")
        return search_query_url

    def _update_url_params(self):
        """retrieves the portions of the url that specify search parameters

        The FHIR API makes union when "where conditions" are requested for neighbouring resources
        Only one "where condition" on every neighbouring resource is taken into account
        """  # noqa
        for resource_alias in self.graph_query.resources_graph.neighbors(self.main_resource_alias):
            edge_info = self.graph_query.resources_graph.edges[
                self.main_resource_alias, resource_alias
            ]["info"]

            resource = self.graph_query.resources_by_alias[resource_alias]
            elements = resource.elements.where(goal="where")

            # Update url for each condition in "where conditions" on resource_alias
            for element in elements:
                search_param = element.search_parameter
                searchparam_prefix = edge_info.searchparam_prefix[self.main_resource_alias]

                self._update_params_dict(search_param, searchparam_prefix=searchparam_prefix)

            logger.debug(f"the part of the url for the params is: {self._params}")

        # "where condition" on the resource itself
        resource_alias_info = self.graph_query.resources_by_alias[self.main_resource_alias]
        elements = resource_alias_info.elements.where(goal="where")

        for element in elements:
            self._update_params_dict(element.search_parameter)

        logger.debug(f"the part of the url for the params is: {self._params}")

    def _update_params_dict(
        self, search_param: Type[SearchParameter], searchparam_prefix: str = ""
    ):
        key = f"{searchparam_prefix}{search_param.code}"
        value = f"{search_param.prefix or ''}{search_param.value}"

        self._params[key].append(value)



import logging
from collections import defaultdict
from pprint import pformat

import networkx as nx

from fhir2dataset.data_class import EdgeInfo, Element, Elements, ResourceAliasInfo, SearchParameter
from fhir2dataset.fhirrules import FHIRRules

logger = logging.getLogger(__name__)


class GraphQuery:
    """Class for storing query information in the form of a graph.

    Attributes:
        resources_graph (nx.Graph): The nodes correspond to the aliases involved in
            the query (filled in the "from") and stops the reference link between 2 aliases
            (filled in the "join").
        resources_by_alias (dict): Dictionary storing information about each alias:
            * the type of the associated resource
            * the elements that must be retrieved from the json of a resource
            * a boolean indicating whether to return the number of instances of the resource
              that meets all these criteria or not
        fhir_rules (FHIRRules): an instance of an FHIRRules object which contains information
            specific to the FHIR standard and the API used (for example the fhirpaths
            associated with the search param of a resource).
    """  # noqa

    def __init__(self, fhir_api_url: str, fhir_rules: FHIRRules) -> None:
        """Instantiate the class and create the query object

        Arguments:
            fhir_api_url (str): the Service Base URL (e.g. http://hapi.fhir.org/baseR4/)
            fhir_rules (FHIRRules): an instance of a FHIRRules-type object
        """  # noqa
        self.fhir_api_url = fhir_api_url
        self.fhir_rules = fhir_rules

        # to represent the relationships (references) between resources
        self.resources_graph = nx.Graph()
        self.resources_by_alias = defaultdict(ResourceAliasInfo)

    def build(
        self, select_dict: dict, from_dict: dict, join_dict: dict = None, where_dict: dict = None
    ):
        """Populates the attributes resources_graph and resources_by_alias according
        to the information filled in.

        Arguments:
            select_dict (dict): elements to be selected from the different resources
            from_dict (dict): all requested resources
            join_dict (dict): the inner join rules between resources (default: None)
            where_dict (dict): the (cumulative) conditions to be met by the resources (default: None)
        """  # noqa
        self._from(**from_dict)
        if join_dict:
            self._join(**join_dict)
        if where_dict:
            self._where(**where_dict)
        self._select(**select_dict)

        # FIXME: Need FHIR2Dataset#96
        # for resource_alias in self.resources_by_alias.keys():
        #     self.resources_by_alias[resource_alias].elements.compute_forest_fhirpaths()

        logger.info(f"The nodes are:{self.resources_graph.nodes()}")
        logger.info("The edges are:")
        logger.info(pformat(list(self.resources_graph.edges(data=True))))
        logger.info("The information gathered for each node is:")
        logger.info(pformat(self.resources_by_alias))

    def _from(self, **resource_type_alias):
        """Initializes the graph nodes contained in resources_graph and the dictionary
        of resources_by_alias information of the aliases listed in resource_type_alias

        Arguments:
            **resource_type_alias: the key corresponds to the alias and the value to the type of the resource.
        """  # noqa
        for resource_alias, resource_type in resource_type_alias.items():
            elements = Elements(
                [
                    Element(
                        col_name="from_id",
                        fhirpath="id",
                        concat_type="row",
                    )
                ]
            )

            self.resources_graph.add_node(resource_alias)

            self.resources_by_alias[resource_alias] = ResourceAliasInfo(
                alias=resource_alias, resource_type=resource_type, elements=elements
            )

    def _join(self, **join_as):
        """Builds the reference links between the aliases involved in the query
        * fills in the elements in attribute resources_by_alias to be retrieved from the
          json resource file to be able to make the joins
        * creates the edges between the nodes of the relevant aliases in the resources_graph
          attribute.

        Keyword Arguments:
            **join_as: the key is an alias of a parent resource and the value is a dictionary
                containing in the key the search parameter leading to the reference, and in the
                value the alias of the child resource.
        """  # noqa

        for join_how, relationships_dict in join_as.items():
            join_how = join_how.lower()
            if join_how not in ["inner", "child", "parent"]:
                raise ValueError("Invalid join: joins should be of type: inner, child or parent")

            for (alias_parent, searchparam_dict) in relationships_dict.items():
                type_parent = self.resources_by_alias[alias_parent].resource_type
                for (searchparam_parent, alias_child) in searchparam_dict.items():
                    # Update element to have in table
                    fhirpath = self.fhir_rules.searchparam_to_fhirpath(
                        resource_type=type_parent, search_param=searchparam_parent
                    )
                    if fhirpath is None:
                        raise ValueError(
                            f"The reference attribute in the JOIN clause should be a valid "
                            f"gaisearchparameter, but got '{searchparam_parent}'."
                        )

                    if " | " in fhirpath:
                        fhirpath_ref = f"(({fhirpath}).reference)"
                    else:
                        fhirpath_ref = f"({fhirpath}.reference)"

                    self.resources_by_alias[alias_parent].elements.append(
                        Element(
                            goal="join",
                            col_name=f"join_{searchparam_parent}",
                            fhirpath=fhirpath_ref,
                            concat_type="row",
                        )
                    )

                    # Update Graph
                    type_child = self.resources_by_alias[alias_child].resource_type

                    searchparam_parent_to_child = f"{searchparam_parent}:{type_child}."
                    # include = f"{type_parent}:{searchparam_parent}:" f"{type_child}"

                    searchparam_child_to_parent = f"_has:{type_parent}:{searchparam_parent}:"
                    # revinclude = f"{type_parent}:{searchparam_parent}:" f"{type_child}"

                    searchparam_prefix = {
                        alias_parent: searchparam_parent_to_child,
                        alias_child: searchparam_child_to_parent,
                    }

                    edge_info = EdgeInfo(
                        parent=alias_parent,
                        child=alias_child,
                        searchparam_parent=searchparam_parent,
                        join_how=join_how,
                        searchparam_prefix=searchparam_prefix,
                    )

                    self.resources_graph.add_edge(alias_parent, alias_child, info=edge_info)

    def _where(self, **wheres):
        """updates the resources_by_alias attribute with the conditions that each alias must meet

        Keyword Arguments:
            **wheres: the key is an alias, the value is a dictionary containing itself keys which
                are searchparams and whose values must be respected by the associated search params
        """  # noqa
        for resource_alias, conditions in wheres.items():
            for search_param, values in conditions.items():
                fhirpath = self._check_searchparam_or_fhirpath(resource_alias, search_param)
                if not isinstance(values, list):
                    values = [values]

                for value in values:
                    prefix = None

                    # handle the case where a prefix has been specified
                    # value_full={"ge": "1970"}
                    if type(value) is dict:
                        for key, val in value.items():
                            prefix = key
                            value = val
                            break

                    search_param_obj = SearchParameter(
                        code=search_param, prefix=prefix, value=value
                    )
                    self.resources_by_alias[resource_alias].elements.append(
                        Element(
                            goal="where",
                            col_name=f"where_{search_param_obj.code}",
                            fhirpath=fhirpath,
                            search_parameter=search_param_obj,
                        )
                    )

    def _select(self, **selects):
        """updates the resources_by_alias attribute with the elements that must be retrieved
        for each alias

        Keyword Arguments:
            **selects: the key is an alias, the value is a list containing fhirpath to leaf
                elements
        """  # noqa
        for resource_alias, col_names in selects.items():
            for col_name in col_names:
                fhirpath = self._check_searchparam_or_fhirpath(resource_alias, col_name)
                self.resources_by_alias[resource_alias].elements.append(
                    Element(
                        goal="select",
                        col_name=col_name,
                        fhirpath=fhirpath,
                    )
                )

    def _check_searchparam_or_fhirpath(self, resource_alias: str, searchparam_or_fhirpath: str):
        """transforms searchparam_or_fhirpath into its fhirpath if it's a searchparam, otherwise it returns the argument as it was entered.

        Args:
            resource_alias (str): alias associated with a resource
            searchparam_or_fhirpath (str): string of characters that can correspond to a searchparam or a fhirpath

        Returns:
            str: string of characters corresponding to a fhirpath
        """  # noqa
        resource_type = self.resources_by_alias[resource_alias].resource_type
        fhirpath = self.fhir_rules.searchparam_to_fhirpath(
            resource_type=resource_type,
            search_param=searchparam_or_fhirpath,
        )
        return fhirpath or searchparam_or_fhirpath



import logging
import multiprocessing
import pprint
from json import JSONDecodeError
from typing import List, Optional

import numpy as np
import pandas as pd
import requests

# from fhir2dataset.fhirpath import fhirpath_processus_tree
from fhir2dataset.data_class import Elements
from fhir2dataset.tools.progressbar import progressbar

logger = logging.getLogger(__name__)

PAGE_SIZE = 100  # Return maximum 300 entities per query


# def concat_cell(dataset, cols_name, element):
#     # column of one row, this single cell is of the same type as element.value, i.e. a list
#     if element.value == []:
#         element.value = [None]
#     column = [element.value]
#     dataset.append(column)
#     cols_name.append(element.col_name)
#     return dataset, cols_name
#
#
# def concat_col(dataset, cols_name, element):
#     for index, value in enumerate(element.value):
#         # column of one row, this single cell is of the same type as value
#         column = [value]
#         dataset.append(column)
#         cols_name.append(f"{element.col_name}_{index}")
#     return dataset, cols_name
#
#
# def concat_row(dataset, cols_name, element):
#     # column of x=len(element.value) rows
#     column = element.value
#     dataset.append(column)
#     cols_name.append(element.col_name)
#     return dataset, cols_name
#
#
# MAPPING_CONCAT = {"cell": concat_cell, "col": concat_col, "row": concat_row}


def process_function(token, url):
    """
    Make a call to a remote API.
    Function called by the different processes when doing parallel querying of the API
    """
    auth = BearerAuth(token)
    response = requests.get(url, auth=auth)
    return response.json()


class Response:
    """class that contains the data retrieves from an url"""

    def __init__(self, total: int = None, results: List = None, next_url: str = None):
        self.total: Optional[int] = total
        self.results: List = results
        self.next_url: Optional[str] = next_url


class BearerAuth(requests.auth.AuthBase):
    def __init__(self, token):
        self.token = token

    def __call__(self, r):
        if self.token:
            r.headers["Authorization"] = self.token
        return r


class ApiCall:
    """Generic class that execute a specific url request to a FHIR API.
    This class also fixes urls that might be broken, and handle errors
    from the API.

    Attributes:
        url (str): the url of the api to call
        auth (BearerAuth): (optional) a bearer toke if necessary
    """  # noqa

    def __init__(self, url: str, token: str = None):
        self.url = url
        self.auth = BearerAuth(token)

    @progressbar
    def _get_response(self, url: str) -> Response:
        logger.info(f"Get {url}")

        url = self.__fix_url(url)

        response = requests.get(url, auth=self.auth)

        failed = False
        try:
            response_content = response.json()
            # One of these entries should be found
            if "entry" not in response_content and "total" not in response_content:
                failed = True
        except JSONDecodeError:
            failed = True

        if not failed:
            results = response_content.get("entry")
            links = response_content.get("link", [])
            next_pages = [link["url"] for link in links if link["relation"] == "next"]

            response = Response(
                total=response_content.get("total", 0),
                results=results,
                next_url=next_pages[0] if next_pages and results else None,
            )
            return response
        else:
            raise ValueError(
                f"Request: {url}\n"
                f"Status code of failing response: {response.status_code}\n"
                f"Content of the failing response:\n{pprint.pformat(response.__dict__)}"
            )

    def _get_count(self, url: str) -> int:
        url_count = f"{url}?_summary=count"
        logger.info(f"Get {url_count}")

        response = self._get_response(url_count)
        return response.total

    def _fix_next_url(self, next_url: str) -> str:
        """Apply a set of fixes for the next_url of the Arkhn Api"""
        # FIXME: Not needed anymore now that we use hapi.
        # FIXME: Remove this function
        # if "arkhn" in next_url:
        #     # Enforce that the base URL is not changed
        #     if self.url not in next_url:
        #         next_url = re.sub(
        #             r"^(?:http:\/\/|www\.|https:\/\/)([^\/]+)",
        #             self.url,
        #             next_url,  # hackalert
        #         )

        if "_count" not in next_url:
            # Append _count info
            next_url = f"{next_url}?_count={PAGE_SIZE}"

        return next_url

    @staticmethod
    def __fix_url(url):
        """Apply a set of hot fixes on the url"""

        # Hot fix: remove noisy "%3D"
        url = url.replace("%3D", "")

        # Hot fix in syntax
        url = url.replace("/?", "?")

        # Hot fix to change several '?' in the url in 1 '?' and '&' after
        # (e.g g.co/api?a=1&b=1?c=1 becomes g.co/api?a=1&b=1&c=1

        seen = False
        clean_url = ""
        for char in url:
            if char == "?":
                if not seen:
                    seen = True
                    clean_url += "?"
                else:
                    clean_url += "&"
            else:
                clean_url += char
        url = clean_url

        return url


class ApiRequest(ApiCall):
    """Class that inherits ApiCall, and manage the Request by splitting it in different call.
    This class manages the organisation of the calls to the api and the processing of the data
    received

    Attributes:
        elements (Elements): instance of the FHIR2Dataset Elements class that allows to list
            all the elements that need to be retrieved from the bundles returned in response to
            the request url
        df (pd.DataFrame): dataframe containing the elements to be recovered in tabular format
        parallel_requests (bool) if true, perform queries in get_all() in parallel using an
            offset functionality of some FHIR Apis
        pbar : tqdm progress bar object
        bar_frac (int): total amount of time allocated to this Api call
    """  # noqa

    def __init__(
        self,
        url: str,
        elements: Elements,
        token: str = None,
        parallel_requests: bool = False,
        pbar=None,
        bar_frac: int = 0,
    ):
        ApiCall.__init__(self, url, token)
        self.elements = elements
        self.df = self._init_data()

        self.parallel_requests = parallel_requests

        self.pbar = pbar
        self.bar_frac = bar_frac
        self.number_calls = None

    def get_all(self):
        """collects all the data corresponding to the initial url request by calling the following pages"""  # noqa
        if self.number_calls is None:
            total_resources = self._get_count(self.url)
            logger.info(f"there are {total_resources} matching resources for {self.url}")
            self.number_calls = int(np.ceil(total_resources / PAGE_SIZE))

        if self.number_calls == 0:
            return self._get_data([])

        if self.parallel_requests:
            urls = []
            for i in range(self.number_calls):
                urls.append(
                    (
                        self.auth.token,
                        f"{self.url}&_getpagesoffset={i*PAGE_SIZE}&_count={PAGE_SIZE}",
                    )
                )

            p = multiprocessing.Pool()
            results = p.starmap(process_function, urls)
            p.close()
            self.pbar.update(self.bar_frac)
        else:
            results = []

            next_url = self.url
            while next_url:
                next_url = self._fix_next_url(next_url)
                response = self._get_response(next_url)
                page_results = self._get_data(response.results)

                results.append(page_results)
                next_url = response.next_url

        self._concat(results)

        return self.df

    def _get_data(self, results: List) -> pd.DataFrame:
        """Retrieves the information from the json instance of a resource that is relevant
        to the query (ie listed in self.elements) and put it in a Dataframe

        Arguments:
            results: list of json resources

        Returns:
            pd.DataFrame: with data extracted from the json resources
        """

        columns = [element.col_name for element in self.elements.elements]
        filtered_resources = []
        for json_resource in results:
            resource = json_resource["resource"]
            data_items = []
            for element in self.elements.elements:
                fhirpath = element.fhirpath.replace("(", "").replace(
                    ")", ""
                )  # TODO: analyze this: breaks .where()
                sub_paths = fhirpath.split(".")
                if len(sub_paths) > 0 and sub_paths[0] == resource["resourceType"]:
                    try:
                        # Try to get recursively the keys from sub_paths[1:]
                        data_item = ApiRequest._rgetattr(resource, sub_paths[1:])
                    except KeyError:
                        data_item = None
                elif fhirpath == "id":
                    data_item = resource["id"]
                else:
                    raise ValueError(f"Invalid fhirpath {fhirpath}")
                data_items.append(data_item)

            filtered_resources.append(data_items)

            # FIXME: Need FHIR2Dataset#96
            # elements = self.elements.elements.copy()
            # data_array = fhirpath_processus_tree(self.elements.forest_dict, resource)
            # for element_value, element in zip(data_array, elements):
            #     element.value = element_value
            # df = self._flatten_item_results(elements)
        df = pd.DataFrame(filtered_resources, columns=columns)

        # Drop duplicate columns (when you have two where clauses on a parameter)
        df = df.loc[:, ~df.columns.duplicated()]

        return df

    def _concat(self, results: List[pd.DataFrame]) -> pd.DataFrame:
        """Recursively concat the results of all the pages together

        Arguments:
            results: the list of the dataframes returned by each call

        Returns:
            pd.Dataframe: a consolidated dataframe containing all the results
        """
        self.df = pd.concat([self.df, *results]).reset_index(drop=True)
        return self.df

    @classmethod
    def _rgetattr(cls, obj, keys):
        """
        Recursively get an element in nested dictionaries

        Example:
            >>> ApiRequest._rgetattr(obj, ['attr1', 'attr2', 'attr3'])
            [Out] obj[attr1][attr2][attr3]

        """
        if not isinstance(obj, list):
            # TODO: Fix because ( ) were removed
            if keys[0].startswith("where"):
                attr, value = keys[0][5:].split("=")
                value = value.replace('"', "").replace("'", "")
                if attr in obj and obj[attr] == value:
                    keys = keys[1:]
                else:
                    return None  # FIXME: nothing should be added to a list of contacts

            if len(keys) == 0:
                return obj
            elif len(keys) == 1:
                return obj[keys[0]]
            else:
                first_key, *keys = keys
                return cls._rgetattr(obj[first_key], keys)
        else:
            return [cls._rgetattr(o, keys) for o in obj]

    # INFO: Disabled as we prefere keep lists for the moment
    # def _flatten_item_results(self, elements: Elements):
    #     """creates the tabular version of the elements given as input argument.
    #     For each element of elements, at least one column is added according to the
    #     following process.
    #     1. The first step is to reproduce the type of concatenation desired for each element
    #     If the concatenation type of the element is:
    #         * cell: a single column is created with a single row. The single cell is therefore
    #           of the same type of element.value, i.e. a list.
    #         * row: a single column is created and creates a row for each element in the
    #           element.value list.
    #         * col: len(element.value) column are created. Each column contains a single cell
    #           composed of an element from the element.value list.
    #
    #     2. The second step is to produce the product of all possible combinations between columns
    #     For example, if at the end of step 1, the table is :
    #     Col_1 | Col_2 | Col_3
    #     pat_1 | Pete  | Ginger
    #     pat_1 | P.    | Ginger
    #           | Peter | G.
    #
    #     The table resulting from step 2 will be :
    #     Col_1 | Col_2 | Col_3
    #     pat_1 | Pete  | Ginger
    #     pat_1 | Pete  | G.
    #     pat_1 | P.    | Ginger
    #     pat_1 | P.    | G.
    #     pat_1 | Peter | Ginger
    #     pat_1 | Peter | G.
    #
    #     Args:
    #         elements (fhir2dataset.Elements): instance of elements
    #
    #     Returns:
    #         pd.DataFrame: resulting dataframe
    #     """  # noqa
    #     cols_name = []
    #     dataset = []
    #
    #     for element in elements:
    #         MAPPING_CONCAT[element.concat_type](dataset, cols_name, element)
    #
    #     df = pd.DataFrame(list(product(*dataset)), columns=cols_name)
    #     return df

    def _init_data(self) -> pd.DataFrame:
        """generation of a dictionary whose keys correspond to the column name and the value to an empty list

        Returns:
            dict -- dictionary described above
        """  # noqa
        data = {}
        for element in self.elements.elements:
            if element.col_name not in data:  # Drop duplicates
                data[element.col_name] = []
        return pd.DataFrame(data)



def progressbar(func):
    """Decorator for class methods that update a progressbar

    The class should have the following attributes:
    - pbar: the tqdm progressbar
    - bar_frac (float): the fraction of this bar allocated to the class instance
    - number_calls (int): the expected number of iterations made in this instance
    """

    def _progressbar(self, *args, **kwargs):
        result = func(self, *args, **kwargs)

        if hasattr(self, "pbar") and self.pbar is not None:
            if hasattr(self, "number_calls") and self.number_calls is not None:
                bar_frac_per_call = self.bar_frac / self.number_calls
                self.pbar.update(bar_frac_per_call)

        return result

    return _progressbar



import pytest

from fhir2dataset.parser import Parser  # noqa


@pytest.mark.parametrize(
    "sql_query",
    [
        "SELECT Patient.name.family FROM Patient",
        "SELECT Patient.name.family FROM Patient;",
        "SELECT p.name.family FROM Patient as p",
        "SELECT p.name.family FROM Patient as p;",
        "SELECT Patient.name.family FROM Patient WHERE Patient.birthdate=1900",
        "SELECT Patient.name.family, Patient.address.city FROM Patient",
        """
SELECT Patient.name.family
FROM Patient
;""",
        """
    SELECT Patient.name.family
      FROM Patient
;""",
    ],
)
def test_select(sql_query):
    Parser().from_sql(sql_query)


def test_from():
    sql_query = "SELECT Patient.name.family FROM Patient AS p"
    with pytest.raises(ValueError):
        Parser().from_sql(sql_query)

    sql_query = "SELECT p.Patient.name.family FROM Patient"
    with pytest.raises(ValueError):
        Parser().from_sql(sql_query)


@pytest.mark.parametrize(
    "sql_query",
    [
        "SELECT p.name.family FROM Patient AS p WHERE p.birthdate=1900-01-01",
        "SELECT p.name.family FROM Patient AS p WHERE p.birthdate =1900-01-01",
        "SELECT p.name.family FROM Patient AS p WHERE p.birthdate= 1900-01-01",
        "SELECT p.name.family FROM Patient AS p WHERE p.birthdate = 1900-01-01",
        "SELECT p.name.family FROM Patient AS p WHERE p.birthdate = 1900-01-01 AND p.gender = 'female'",  # noqa
        """
    SELECT
    Patient.gender
    From Patient
    where Patient.birthdate=ge2000-01-01 and Patient.birthdate=le2000-01-31""",
    ],
)
def test_where(sql_query):
    Parser().from_sql(sql_query)



import json
import logging
import os

import pytest

from fhir2dataset.query import Query
from tests.tools import create_resource_test

log_format = "[%(asctime)s] [%(levelname)s] - %(message)s"
logging.basicConfig(level=logging.INFO, format=log_format)


@pytest.mark.parametrize(
    "dirname, fhir_api_url",
    [
        ("tests/1", "http://hapi.fhir.org/baseR4/"),
        # ("tests/2", "http://hapi.fhir.org/baseR4/"),
        ("tests/3", "http://hapi.fhir.org/baseR4/"),
        # ("tests/4", "http://hapi.fhir.org/baseR4/"),
        ("tests/5", "http://hapi.fhir.org/baseR4/"),
    ],
)
def test_resources_in_dataframe(dirname, fhir_api_url):
    create_resource_test(dirname, fhir_api_url)
    with open(os.path.join(dirname, "config.json")) as json_file:
        config = json.load(json_file)
    with open(os.path.join(dirname, "infos_test", "config_checks.json")) as json_file:
        checks = json.load(json_file)
    with open(os.path.join(dirname, "infos_test", "info_hapi.json")) as json_file:
        info_hapi = json.load(json_file)

    query = Query(fhir_api_url)
    query.from_config(config)
    query.execute(debug=True)
    df = query.main_dataframe

    lines = checks["line"]
    for line in lines:
        cols = []
        conditions = []
        for alias, filename in line:
            cols.append(f"{alias}:from_id")
            conditions.append(info_hapi[filename])
        condition = " & ".join(
            [f"(df['{col}'].str.contains('{cond}'))" for col, cond in zip(cols, conditions)]
        )
        logging.info(condition)
        result = df[eval(condition)]
        logging.info(result)
        assert len(result.index) >= 1, f"{dirname} failed"



"""classes representing the different types of information manipulated in FHIR2Dataset
"""
import logging
from dataclasses import dataclass, field
from typing import List, Optional

from fhir2dataset.tools.visualization import custom_repr

logger = logging.getLogger(__name__)


@dataclass
class SearchParameter:
    code: Optional[str] = field(default=None)
    fhirpath: Optional[str] = field(default=None)
    resource_types: Optional[List[str]] = field(default=None)
    prefix: Optional[str] = field(default=None)
    value: Optional[str] = field(default=None)


@dataclass
class Element:
    """unit entity that can be found in a fhir instance in json format using fhirpath or in a table in the col_name column.

    An element is destined to be associated with a single value. For example, if we want to retrieve the patient id of a bundle containing 2 patients, we will have to create 2 element instances.
    """  # noqa

    col_name: str
    fhirpath: str
    goal: str = field(default="select")  # select, where or join
    concat_type: Optional[str] = field(default="cell")
    search_parameter: Optional["SearchParameter"] = field(default=None)


@dataclass
class Elements:
    """collection allowing to group together elements"""

    elements: List[Element] = field(default_factory=list)

    def append(self, x):
        self.elements.append(x)

    def where(self, goal):
        return [element for element in self.elements if element.goal == goal]

    # FIXME: Need FHIR2Dataset#96
    # def compute_forest_fhirpaths(self):
    #     forest = Forest()
    #     for fhirpath in [element.fhirpath for element in self.elements]:
    #         forest.add_fhirpath(fhirpath)
    #     forest.simplify_trees()
    #     forest.parse_fhirpaths()
    #     self.forest = forest
    #     self.forest_dict = forest.create_forest_dict()


@dataclass
class ResourceAliasInfoBasic:
    alias: str
    resource_type: str
    elements: Elements() = field(default_factory=Elements())


class ResourceAliasInfo(ResourceAliasInfoBasic):
    def __repr__(self):
        return custom_repr(super().__repr__())


@dataclass
class EdgeInfo:
    parent: str
    child: str
    searchparam_parent: Optional[str] = field(default=None)
    join_how: str = field(default="inner")
    searchparam_prefix: Optional[dict] = field(default=None)
